from __future__ import annotations

import asyncio

import pytest

import luxos.asyncops as aapi
from luxos import exceptions, misc

## NOTE ##
# This tests spawn an underlying server, it might be better not run
# unattended. Some also require a miner, we might not have it handy.
pytestmark = pytest.mark.manual


def test_parameters_to_list():
    assert aapi.parameters_to_list(None) == []
    assert aapi.parameters_to_list(0) == ["0"]
    assert aapi.parameters_to_list([0]) == ["0"]
    assert aapi.parameters_to_list(["0"]) == ["0"]
    assert aapi.parameters_to_list("hello") == ["hello"]
    assert aapi.parameters_to_list(["hello", "world"]) == ["hello", "world"]
    assert aapi.parameters_to_list(["hello", 1]) == ["hello", "1"]
    assert aapi.parameters_to_list({"hello": 1}) == ["hello=1"]
    assert aapi.parameters_to_list({"hello": True}) == ["hello=true"]


def test_validate_check_syntax(resolver):
    """validate messages malformed (eg. missing STATUS or id"""
    pytest.raises(exceptions.MinerMessageMalformedError, aapi.validate, {})
    pytest.raises(exceptions.MinerMessageMalformedError, aapi.validate, {"id": 1})
    pytest.raises(exceptions.MinerMessageMalformedError, aapi.validate, {"STATUS": 1})
    assert aapi.validate({"STATUS": 0, "id": 1}) == {"STATUS": 0, "id": 1}
    assert aapi.validate({"STATUS": 0, "id": 1}, None, 1000) == {"STATUS": 0, "id": 1}
    pytest.raises(RuntimeError, aapi.validate, None, "GROUPS", 3, 1)


def test_validate_with_extrakey_check_one_result(resolver):
    """validate message like version with a single result"""
    res = resolver.load("messages/version.json")

    # there's been an error
    res["STATUS"][0]["STATUS"] = "E"
    pytest.raises(exceptions.MinerMessageError, aapi.validate, res, "NOT-KEY", 1)

    # status is ok
    res["STATUS"][0]["STATUS"] = "S"

    # we dont' expect results (eg. minfields 0), some messages as POOLS
    # may not return any item in the 'POLLS' key if there aren't pools
    assert aapi.validate(res, "NOT-KEY", 0) is None
    assert aapi.validate(res, "NOT-KEY", None) is None

    # we do expect at least one item
    pytest.raises(exceptions.MinerMessageInvalidError, aapi.validate, res, "NOT-KEY", 1)
    assert aapi.validate(res, "VERSION", 1, None) == [
        {
            "API": "3.7",
            "CompileTime": "Mon Jan 01 00:01:01 UTC 2000",
            "LUXminer": "2000.1.1.12345-aaabbbc",
            "Miner": "some id",
            "Type": "A model",
        }
    ]

    # we do expect exactly one item, we'll return it (list vs item)
    assert aapi.validate(res, "VERSION", 1, 1) == {
        "API": "3.7",
        "CompileTime": "Mon Jan 01 00:01:01 UTC 2000",
        "LUXminer": "2000.1.1.12345-aaabbbc",
        "Miner": "some id",
        "Type": "A model",
    }


def test_validate_with_extrakey_check_many_results(resolver):
    """validate message like groups with multiple results"""
    res = resolver.load("messages/groups.json")

    # there's been an error
    res["STATUS"][0]["STATUS"] = "E"
    pytest.raises(exceptions.MinerMessageError, aapi.validate, res, "NOT-KEY", 1)

    # status is ok
    res["STATUS"][0]["STATUS"] = "S"

    # we dont' expect results (eg. minfields 0), some messages as POOLS
    # may not return any item in the 'POLLS' key if there aren't pools
    assert aapi.validate(res, "NOT-KEY", 0) is None
    assert aapi.validate(res, "NOT-KEY", None) is None

    # we do expect at least one item
    pytest.raises(exceptions.MinerMessageInvalidError, aapi.validate, res, "NOT-KEY", 1)

    pytest.raises(
        exceptions.MinerMessageInvalidError, aapi.validate, res, "GROUPS", None, 1
    )
    pytest.raises(
        exceptions.MinerMessageInvalidError, aapi.validate, res, "GROUPS", 1, 2
    )
    pytest.raises(
        exceptions.MinerMessageInvalidError, aapi.validate, res, "GROUPS", 1, 3
    )
    groups = aapi.validate(res, "GROUPS", 1, 4)
    groups1 = aapi.validate(res, "GROUPS", 1, None)
    assert groups == groups1
    assert isinstance(groups, list) and len(groups) == 4
    assert groups[3]["GROUP"] == 3 and groups[3]["Name"] == "test-group2"
    pytest.raises(
        exceptions.MinerMessageInvalidError, aapi.validate, res, "GROUPS", 99, 100
    )
    pytest.raises(
        exceptions.MinerMessageInvalidError, aapi.validate, res, "GROUPS", 99, None
    )


def test_validate_message():
    # TODO add more cases
    # 1. when res is not present and limit is 0,None
    # 2. check limits like None/value value/None None/None

    # Avoid to repeat the same code over and over
    def validate(res, extrakey=None, minfields=1, maxfields=1):
        return aapi.validate_message("a-host", 0, res, extrakey, minfields, maxfields)

    res = {}
    pytest.raises(exceptions.MinerCommandMalformedMessageError, validate, res)

    res = {"STATUS": [{"STATUS": "S"}], "id": 2}
    assert validate(res)
    assert validate(res, "A-MISSING-KEY", 0, None) == []
    assert validate(res, "A-MISSING-KEY", None, None) == []

    pytest.raises(
        exceptions.MinerCommandMalformedMessageError, validate, res, "A-MISSING-KEY"
    )
    pytest.raises(
        exceptions.MinerCommandMalformedMessageError,
        validate,
        res,
        "A-MISSING-KEY",
        1,
        None,
    )

    res = {"STATUS": [{"STATUS": "S"}], "id": 2, "KEY": [1, 2, 3]}
    assert validate(res)
    assert validate(res, "KEY", 0, None)["KEY"] == [1, 2, 3]
    assert validate(res, "KEY", None, None)["KEY"] == [1, 2, 3]
    assert validate(res, "KEY", None, 5)["KEY"] == [1, 2, 3]
    assert validate(res, "KEY", 1, 5)["KEY"] == [1, 2, 3]

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        validate(res, "KEY", 4, None)
    assert excinfo.value.args[2] == "found too few items for 'KEY'  (3 < 4)"

    with pytest.raises(exceptions.MinerCommandMalformedMessageError) as excinfo:
        validate(res, "KEY", 1, 2)
    assert excinfo.value.args[2] == "found too many items for 'KEY'  (3 > 2)"


@pytest.mark.asyncio
async def test_private_roundtrip_one_listener(echopool):
    """checks roundrtip sends and receive a message (1-listener)"""
    echopool.start(1, mode="echo+")
    host, port = echopool.addresses[0]
    ret = await aapi._roundtrip(host, port, "hello", None)
    assert ret == f"received by ('{host}', {port}): hello"


@pytest.mark.asyncio
async def test_private_roundtrip_many_listeners(echopool):
    """checks the roundrip can connect en-masse to many lsiteners"""
    echopool.start(100, mode="echo+")

    blocks = {}
    for index, group in enumerate(misc.batched(echopool.addresses, 10)):
        tasks = []
        for host, port in group:
            tasks.append(aapi._roundtrip(host, port, "hello olleh", None))
        blocks[index] = await asyncio.gather(*tasks, return_exceptions=True)

    assert len(blocks) == 10
    assert sum(len(x) for x in blocks.values()) == 100
    allitems = [item for b in blocks.values() for item in b]
    assert f"received by {echopool.addresses[3]}: hello olleh" in allitems


@pytest.mark.asyncio
async def test_miner_logon_logoff_cycle(miner_host_port):
    host, port = miner_host_port
    await aapi.rexec(host, port, "kill")

    sid = None
    try:
        sid = await aapi.logon(host, port)
    except exceptions.MinerCommandSessionAlreadyActive as e:
        raise RuntimeError("a session is already active on {host}:{port}") from e
    finally:
        if sid:
            await aapi.logoff(host, port, sid)


@pytest.mark.asyncio
async def test_miner_double_logon_cycle(miner_host_port):
    host, port = miner_host_port

    await aapi.rexec(host, port, "kill")

    sid = await aapi.logon(host, port)
    try:
        with pytest.raises(exceptions.MinerCommandSessionAlreadyActive) as excinfo:
            await aapi.logon(host, port)
        assert excinfo.value.args[0] == f"session active for {host}:{port}"
    except Exception:
        pass
    finally:
        if sid:
            await aapi.logoff(host, port, sid)


@pytest.mark.asyncio
async def test_miner_version(miner_host_port):
    host, port = miner_host_port

    res = await aapi.rexec(host, port, "version")
    assert "VERSION" in res
    assert len(res["VERSION"]) == 1
    assert "API" in res["VERSION"][0]


@pytest.mark.asyncio
async def test_miner_profile_sets(miner_host_port):
    from random import choices
    from string import ascii_lowercase

    # random profile name
    profile = f"test-{''.join(choices(ascii_lowercase, k=5))}"

    # get the initial profile list
    host, port = miner_host_port
    profiles = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
    assert profiles and profile not in {p["Profile Name"] for p in profiles}

    # create a new profile
    # TODO there's a bug
    #  when ATM is running you shouldn't be able to create/delete
    #  profiles, the following test should fail at the first try
    # -> to pass this test, you need to disable ATM

    params = f"{profile},{profiles[0]['Frequency']},{profiles[0]['Voltage']}"
    await aapi.rexec(host, port, "profilenew", params)

    try:
        profiles1 = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
        assert profile in {p["Profile Name"] for p in profiles1}
    finally:
        async with aapi.with_atm(host, port, False):
            await aapi.rexec(host, port, "profilerem", profile)

    # verify we restored the same profiles
    profiles2 = (await aapi.rexec(host, port, "profiles"))["PROFILES"]
    expected = {p["Profile Name"] for p in profiles}
    found = {p["Profile Name"] for p in profiles2}
    assert found == expected


@pytest.mark.asyncio
async def test_roundtrip_timeout(miner_host_port):
    # host, port = miner
    """checks roundrtip sends and receive a message (1-listener)"""

    # miner doesn't exist
    host, port = "127.0.0.99", 12345
    try:
        await aapi.rexec(host, port, "hello", timeout=0.5)
        assert False, "didn't raise"
    except aapi.exceptions.MinerCommandTimeoutError as exc:
        exception = exc
    assert isinstance(exception, asyncio.TimeoutError)
    assert isinstance(exception, aapi.exceptions.MinerConnectionError)
    assert isinstance(exception, aapi.exceptions.LuxosBaseException)
    text = f"<{host}:{port}>: MinerCommandTimeoutError, TimeoutError()"
    assert str(exception)[: len(text)] == text

    # not miner on a wrong port
    host, port = miner_host_port
    port = 22
    exception = None
    try:
        await aapi.rexec(host, port, "hello", timeout=0.5)
        assert False, "didn't raise"
    except aapi.exceptions.MinerCommandTimeoutError as exc:
        exception = exc
    assert isinstance(exception, asyncio.TimeoutError)
    assert isinstance(exception, aapi.exceptions.MinerConnectionError)
    assert isinstance(exception, aapi.exceptions.LuxosBaseException)
    texts = [
        f"<{host}:{port}>: MinerCommandTimeoutError, ConnectionRe",
        f"<{host}:{port}>: MinerCommandTimeoutError, TimeoutError",
        # TODO refactor wire errors
        f"<{host}:{port}>: MinerCommandTimeoutError, JSONDecodeEr",
    ]
    assert str(exception)[: len(texts[0])] in texts


@pytest.mark.asyncio
async def test_bridge_execute_command(miner_host_port):
    from luxos.utils import execute_command, rexec

    # get the initial profile list
    host, port = miner_host_port

    out = execute_command(host, port, 3, "profiles", parameters=[], verbose=True)
    out1 = await rexec(host, port, cmd="profiles")
    assert out["PROFILES"] == out1["PROFILES"]

    port += 1
    pytest.raises(
        exceptions.MinerCommandTimeoutError,
        execute_command,
        host,
        port,
        3,
        "profiles",
        parameters=[],
        verbose=True,
    )



## NOTE ##
# Some of these tests spawn an underlying server, it might be better not run
# unattended. Some also require a miner, we might not have it handy.
import asyncio

import pytest

import luxos.asyncops
from luxos import utils


@pytest.mark.manual
@pytest.mark.asyncio
async def test_basic_call(echopool):
    """places few round trip calls"""
    echopool.start(1)
    host, port = echopool.addresses[0]

    # this will bomb at around 10k
    for index in range(1_000):
        ret = await luxos.asyncops.roundtrip(
            host, port, f"hello world {index}", asjson=False
        )
        assert ret == f"received by ('{host}', {port}): hello world {index}"


@pytest.mark.manual
@pytest.mark.asyncio
async def test_basic_batched_call(echopool):
    """places few round trip calls"""
    echopool.start(1)
    host, port = echopool.addresses[0]

    # this will bomb at around 10k if run sequentially!
    calls = [
        luxos.asyncops.roundtrip(host, port, f"hello world {index}", asjson=False)
        for index in range(10_000)
    ]
    expected = {
        f"received by ('{host}', {port}): hello world {index}"
        for index in range(len(calls))
    }
    found = set()
    from luxos.misc import batched

    for group in batched(calls, 100):
        tasks = await asyncio.gather(*group, return_exceptions=True)
        found.update(tasks)
    assert expected == found


@pytest.mark.manual
@pytest.mark.asyncio
async def test_util_launch_oneshot(echopool):
    """places few round trip calls"""
    echopool.start(1)
    host, port = echopool.addresses[0]

    async def call(index, *_):
        return await luxos.asyncops.roundtrip(
            host, port, f"hello world {index}", asjson=False
        )

    addresses = [(index, None) for index in range(10)]

    expected = {
        f"received by ('{host}', {port}): hello world {index}" for index, _ in addresses
    }

    ret = await utils.launch(addresses, call)
    assert set(ret) == expected


@pytest.mark.manual
@pytest.mark.asyncio
async def test_util_launch_batched(echopool):
    """places *many* round trip calls"""

    echopool.start(1)
    host, port = echopool.addresses[0]

    async def call(index, *_):
        return await luxos.asyncops.roundtrip(
            host, port, f"hello world {index}", asjson=False
        )

    # utils.launch will choke at 10k without batch argument
    addresses = [(index, None) for index in range(10_000)]
    expected = {
        f"received by ('{host}', {port}): hello world {index}" for index, _ in addresses
    }

    ret = await utils.launch(addresses, call, batch=100)
    assert set(ret) == expected


def test_util_ip_ranges():
    assert set(utils.ip_ranges("127.0.0.1")) == {("127.0.0.1", None)}
    assert set(utils.ip_ranges("127.0.0.1-127.0.0.3", gsep="\n")) == {
        ("127.0.0.1", None),
        ("127.0.0.2", None),
        ("127.0.0.3", None),
    }
    assert set(utils.ip_ranges("127.0.0.1-127.0.0.3\n127.0.0.15", gsep="\n")) == {
        ("127.0.0.1", None),
        ("127.0.0.2", None),
        ("127.0.0.3", None),
        ("127.0.0.15", None),
    }


@pytest.mark.asyncio
@pytest.mark.manual
async def test_launch_with_exceptions(miner_host_port):
    host, port = miner_host_port

    async def broken(host, port):
        return (await utils.rexec(host, port, "config"))["CONFIG"][0]["ProfilXe"]

    result = (await utils.launch([(host, port)], broken, asobj=True))[0]
    assert isinstance(result, utils.LuxosLaunchBaseResult)
    assert isinstance(result, utils.LuxosLaunchError)
    assert "KeyError: 'ProfilXe'" in result.traceback

    result = (await utils.launch([(host, port + 1)], broken))[0]
    assert isinstance(result, utils.LuxosLaunchTimeoutError)
    assert isinstance(result, asyncio.TimeoutError)
    assert "ConnectionRefusedError" in str(result.traceback)



# catch-all module (find later a better place)
from __future__ import annotations

import ipaddress
import itertools
import sys
import types
from pathlib import Path
from typing import Generator

if sys.version_info >= (3, 12):
    batched = itertools.batched
else:

    def batched(iterable, n):
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(itertools.islice(it, n)):
            yield batch


def iter_ip_ranges(
    txt: str, gsep: str = ":", rsep: str = "-"
) -> Generator[str, None, None]:
    """iterate over ip ranges

    for ip in iter_ip_ranges("127.0.0.1-127.0.0.3:127.0.0.15"):
        print(ip)

    127.0.0.1
    127.0.0.2
    127.0.0.3
    127.0.0.15

    NOTE: use the `:` (gsep) to separate ips groups, and `-` (rsep) to define a range.
    """

    for segment in txt.split(gsep):
        start, _, end = segment.partition(rsep)
        if not end:
            yield start
        else:
            cur = ipaddress.IPv4Address(start)
            last = ipaddress.IPv4Address(end)
            while cur <= last:
                yield str(cur)
                cur += 1


def loadmod(path: Path) -> types.ModuleType:
    from importlib import util

    spec = util.spec_from_file_location(Path(path).name, Path(path))
    module = util.module_from_spec(spec)  # type: ignore
    spec.loader.exec_module(module)  # type: ignore
    return module



# ruff: noqa: W291, W293
from __future__ import annotations

from types import ModuleType

import pytest

from luxos import text

rich: ModuleType | None = None
try:
    import rich
except ModuleNotFoundError:
    pass


def test_indent():
    txt = """

            Lorem Ipsum is simply dummy text of the printing and
          typesetting industry. Lorem Ipsum has been the industry's standard
         dummy text ever since the 1500s, when an unknown printer
           took a galley of type and scrambled it to make a type specimen book.
"""

    assert (
        text.indent(txt, "." * 2)
        == """\
..
..
..   Lorem Ipsum is simply dummy text of the printing and
.. typesetting industry. Lorem Ipsum has been the industry's standard
..dummy text ever since the 1500s, when an unknown printer
..  took a galley of type and scrambled it to make a type specimen book.
"""
    )


def test_indent2():
    txt = """\
     An unusually complicated text
    with un-even indented lines
   that make life harder
"""
    assert (
        text.indent(txt, pre="..")
        == """\
..  An unusually complicated text
.. with un-even indented lines
..that make life harder
"""
    )


def test_md():
    txt = """
    ### An example

    This is an example of help file, written in MD. The text.md(txt) function should
    format this nicely:
    
    - item 1
    - item 2
    
    | Tables   |      Are      |  Cool |
    |----------|:-------------:|------:|
    | col 1 is |  left-aligned | $1600 |
    | col 2 is |    centered   |   $12 |
    | col 3 is | right-aligned |    $1 |
"""

    assert (
        text.md(txt, md=False)
        == """
### An example

This is an example of help file, written in MD. The text.md(txt) function should
format this nicely:

- item 1
- item 2

| Tables   |      Are      |  Cool |
|----------|:-------------:|------:|
| col 1 is |  left-aligned | $1600 |
| col 2 is |    centered   |   $12 |
| col 3 is | right-aligned |    $1 |
"""
    )


@pytest.mark.skipif(not rich, reason="need rich installed")
def test_md_rich():
    txt = """
        ### An example

        This is an example of help file, written in MD. The text.md(txt) function should
        format this nicely:

        - item 1
        - item 2

        | Tables   |      Are      |  Cool |
        |----------|:-------------:|------:|
        | col 1 is |  left-aligned | $1600 |
        | col 2 is |    centered   |   $12 |
        | col 3 is | right-aligned |    $1 |
    """

    assert (
        text.md(txt, width=80)
        == """\
                                   An example                                   

This is an example of help file, written in MD. The text.md(txt) function should
format this nicely:                                                             

 • item 1                                                                       
 • item 2                                                                       

                                    
  Tables          Are         Cool  
 ━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━━ 
  col 1 is   left-aligned    $1600  
  col 2 is     centered        $12  
  col 3 is   right-aligned      $1  
                                    
"""
    )
    pass



from string import ascii_lowercase

import pytest

from luxos import exceptions, syncops

pytestmark = pytest.mark.manual


def test__roundtrip(host, port):
    pytest.raises(ConnectionRefusedError, syncops._roundtrip, host, port + 1, "version")
    res = syncops._roundtrip(host, port, "version")
    assert res.startswith("STATUS=S")


def test_roundtrip(host, port):
    pytest.raises(
        exceptions.MinerCommandTimeoutError,
        syncops.roundtrip,
        host,
        port + 1,
        "version",
        retry=2,
    )

    res = syncops.roundtrip(host, port, "version", asjson=False)
    assert res.startswith("STATUS=S")


def test_miner_logon_logoff_cycle(miner_host_port):
    host, port = miner_host_port

    res = syncops.roundtrip(host, port, "kill", asjson=False)
    assert res.startswith("STATUS=S")

    sid = None
    try:
        sid = syncops.logon(host, port)
    except exceptions.MinerCommandSessionAlreadyActive as e:
        raise RuntimeError("a session is already active on {host}:{port}") from e
    finally:
        if sid:
            syncops.logoff(host, port, sid)


def test_miner_double_logon_cycle(miner_host_port):
    host, port = miner_host_port

    res = syncops.roundtrip(host, port, "kill", asjson=False)
    assert res.startswith("STATUS=S")

    sid = syncops.logon(host, port)
    try:
        with pytest.raises(exceptions.MinerCommandSessionAlreadyActive) as excinfo:
            syncops.logon(host, port)
        assert excinfo.value.args[0] == f"session active for {host}:{port}"
    except Exception:
        pass
    finally:
        if sid:
            syncops.logoff(host, port, sid)


def test_miner_version(miner_host_port):
    host, port = miner_host_port

    res = syncops.rexec(host, port, "version")
    assert "VERSION" in res
    assert len(res["VERSION"]) == 1
    assert "API" in res["VERSION"][0]


def test_miner_profile_sets(miner_host_port):
    from random import choices

    # random profile name
    profile = f"test-{''.join(choices(ascii_lowercase, k=5))}"

    # get the initial profile list
    host, port = miner_host_port
    profiles = (syncops.rexec(host, port, "profiles"))["PROFILES"]
    assert profiles and profile not in {p["Profile Name"] for p in profiles}

    # create a new profile
    # TODO there's a bug
    #  when ATM is running you shouldn't be able to create/delete
    #  profiles, the following test should fail at the first try
    # -> to pass this test, you need to disable ATM

    params = f"{profile},{profiles[0]['Frequency']},{profiles[0]['Voltage']}"
    syncops.rexec(host, port, "profilenew", params)

    try:
        profiles1 = (syncops.rexec(host, port, "profiles"))["PROFILES"]
        assert profile in {p["Profile Name"] for p in profiles1}
    finally:
        with syncops.with_atm(host, port, False):
            syncops.rexec(host, port, "profilerem", profile)

    # verify we restored the same profiles
    profiles2 = (syncops.rexec(host, port, "profiles"))["PROFILES"]
    expected = {p["Profile Name"] for p in profiles}
    found = {p["Profile Name"] for p in profiles2}
    assert found == expected


def test_roundtrip_timeout(miner_host_port):
    # host, port = miner
    """checks roundrtip sends and receive a message (1-listener)"""

    # miner doesn't exist
    host, port = "127.0.0.99", 12345
    try:
        syncops.rexec(host, port, "hello", timeout=0.5)
        assert False, "didn't raise"
    except exceptions.MinerCommandTimeoutError as exc:
        exception = exc
    assert isinstance(exception, TimeoutError)
    assert isinstance(exception, exceptions.MinerConnectionError)
    assert isinstance(exception, exceptions.LuxosBaseException)
    text = f"<{host}:{port}>: MinerCommandTimeoutError, TimeoutError("
    assert str(exception)[: len(text)] == text

    # not miner on a wrong port
    host, port = miner_host_port
    port = 22
    exception = None
    try:
        syncops.rexec(host, port, "hello", timeout=0.5)
        assert False, "didn't raise"
    except exceptions.MinerCommandTimeoutError as exc:
        exception = exc
    assert isinstance(exception, TimeoutError)
    assert isinstance(exception, exceptions.MinerConnectionError)
    assert isinstance(exception, exceptions.LuxosBaseException)
    texts = [
        f"<{host}:{port}>: MinerCommandTimeoutError, ConnectionRe",
        f"<{host}:{port}>: MinerCommandTimeoutError, TimeoutError",
    ]
    assert str(exception)[: len(texts[0])] in texts


def test_atm_flip(miner_host_port):
    host, port = miner_host_port

    def getatm():
        res = syncops.rexec(host, port, "atm")
        return syncops.validate_message(host, port, res, "ATM")["ATM"][0]["Enabled"]

    # current status
    status = getatm()
    assert status in {True, False}

    # check we can set the atm to the same status
    syncops.rexec(host, port, "atmset", {"enabled": status})
    assert status == getatm()

    syncops.rexec(host, port, "atmset", {"enabled": not status})
    assert status != getatm()

    syncops.rexec(host, port, "atmset", {"enabled": not getatm()})
    assert status == getatm()



"""cli utilities

This is the v1.cli decorator to help writing cli scripts with a consistent
interface.

A simple v1 script will always have:
* `-v/--verbose | -q/--quiet` flags to increase the logging verbosity level
* `-c/--config` to pass a config file (default to config.yaml)

A `sample.py` script with default sensible and consistent interface:

```
import argparse
import luxos.cli.v1 as cli
import logging

log = logging.getLogger(__name__)


@cli.cli()
def main(args: argparse.Namespace):
    log.debug("a debug message, need to use -v|--verbose to display it")
    log.info("an info message, you can silence it with -q|--quiet")
    log.warning("a warning!")


if __name__ == "__main__":
    main()
```


### Advanced usages

#### changing the default config file

In the `sample.py` file, just add:

```
CONFIGPATH = "my.config.file.yaml"
```

`CONFIGPATH` is part of the module level "magic" variables.

Another one is `LOGGING_CONFIG` (not recommended!):
```
LOGGING_CONFIG = {
    'level': logging.INFO,
    'format': "%(asctime)s:%(name)s:[%(levelname)s] %(message)s",
    'handlers': [
        logging.StreamHandler(),
        logging.FileHandler("LuxOS-LoadControl.log")
    ],
}
```

The `LOGGING_CONFIG` is a dictionary feed into `logging.basicConfig(**LOGGING_CONFIG)`.

#### add and process new extra arguments

in the `sample.py` file:

```
def add_arguments(parser: argparse.ArgumentParser) -> None:
    ... adds as many args needed


def process_args(args: argparse.Namespace) -> argparse.Namespace | None:
    ... you can manipulate args in place, changing its attributes in place
    ... if you return a new Namespace instance then it will be fee to main below


@cli.cli(add_arguments, process_args)
def main(args: argparse.Namespace):
    ... args is the Namespace instance returned by process_args if defined,
    ... or the default processed one of process_args is not provided

```

#### escape hatch to handle all the parsing/processing

```

@cli.cli()
def main(parser: argparse.ArgumentParser):
    ... you can use directly the parser here

```

"""

from __future__ import annotations

import argparse
import contextlib
import functools
import inspect
import logging.handlers
import sys
import time
import types
from pathlib import Path
from typing import Any, Callable

from .. import text
from . import flags
from .shared import ArgumentTypeBase, LuxosParserBase


class MyHandler(logging.StreamHandler):
    def emit(self, record):
        record.shortname = record.levelname[0]
        return super().emit(record)


# SPECIAL MODULE LEVEL VARIABLES
LOGGING_CONFIG = {
    "format": "%(asctime)s [%(shortname)s] %(name)s: %(message)s",
    "handlers": [
        MyHandler(),
    ],
}
CONFIGPATH = Path("config.yaml")


log = logging.getLogger(__name__)


class CliBaseError(Exception):
    pass


class AbortCliError(CliBaseError):
    pass


class AbortWrongArgumentError(CliBaseError):
    pass


def log_sys_info():
    from luxos import __hash__, __version__

    log.info(
        "py[%s], luxos[%s/%s]", sys.version.partition(" ")[0], __version__, __hash__
    )
    log.debug("interpreter: %s", sys.executable)
    log.debug("version: %s", sys.version)


ArgumentParser = LuxosParserBase


class LuxosParser(LuxosParserBase):
    def __init__(self, modules: list[types.ModuleType], *args, **kwargs):
        super().__init__(modules, *args, **kwargs)

        # we're adding the -v|-q flags, to control the logging level
        flags.add_arguments_logging(self)

    def error(self, message: str):
        raise AbortWrongArgumentError(message)

    def parse_args(self, args=None, namespace=None):
        options = super().parse_args(args, namespace)

        # reserver attributes
        for reserved in [
            "modules",
            "error",
        ]:
            if not hasattr(options, reserved):
                continue
            raise RuntimeError(f"cannot add an argument with dest='{reserved}'")
        options.error = self.error
        options.modules = self.modules

        for name in dir(options):
            if isinstance(getattr(options, name), ArgumentTypeBase):
                setattr(options, name, getattr(options, name).value)

        for callback in self.callbacks:
            if not callback:
                continue
            options = callback(options) or options

        log_sys_info()
        return options

    @classmethod
    def get_parser(cls, modules: list[types.ModuleType], **kwargs):
        class Formatter(
            argparse.ArgumentDefaultsHelpFormatter, argparse.RawTextHelpFormatter
        ):
            pass

        return cls(modules, formatter_class=Formatter, **kwargs)


@contextlib.contextmanager
def setup(
    function: Callable,
    add_arguments: (
        Callable[[LuxosParserBase], None]
        | Callable[[argparse.ArgumentParser], None]
        | None
    ) = None,
    process_args: (
        Callable[[argparse.Namespace], argparse.Namespace | None] | None
    ) = None,
):
    sig = inspect.signature(function)
    module = inspect.getmodule(function)

    if "args" in sig.parameters and "parser" in sig.parameters:
        raise RuntimeError("cannot use args and parser at the same time")

    description, _, epilog = (
        (function.__doc__ or module.__doc__ or "").strip().partition("\n")
    )
    # markdown
    # epilog = text.md(f"# {description}\n{epilog}")
    epilog = text.md(f"{description}\n{epilog}")
    description = ""

    kwargs = {}
    modules = [
        sys.modules[__name__],
    ]
    if module:
        modules.append(module)
    parser = LuxosParser.get_parser(modules, description=description, epilog=epilog)
    if add_arguments and (callbacks := add_arguments(parser)):
        if isinstance(callbacks, list):
            parser.callbacks.extend(callbacks)
        else:
            parser.callbacks.append(callbacks)

    if "parser" in sig.parameters:
        kwargs["parser"] = parser

    t0 = time.monotonic()
    success = "completed"
    errormsg = ""
    show_timing = True
    try:
        if "parser" not in sig.parameters:
            args = parser.parse_args()
            if process_args:
                args = process_args(args) or args

            if "args" in sig.parameters:
                kwargs["args"] = args
        yield sig.bind(**kwargs)
    except AbortCliError as exc:
        show_timing = False
        if exc.args:
            print(str(exc), file=sys.stderr)
        sys.exit(2)
    except AbortWrongArgumentError as exc:
        show_timing = False
        parser.print_usage(sys.stderr)
        print(f"{parser.prog}: error: {exc.args[0]}", file=sys.stderr)
        sys.exit(2)
    except SystemExit as exc:
        show_timing = False
        sys.exit(exc.code)
    except Exception:
        log.exception("un-handled exception")
        success = "failed"
    finally:
        if show_timing:
            delta = round(time.monotonic() - t0, 2)
            log.info("task %s in %.2fs", success, delta)
    if errormsg:
        parser.error(errormsg)


def cli(
    # add_arguments: Callable[[LuxosParserBase | argparse.ArgumentParser], Any]
    add_arguments: (
        Callable[[LuxosParserBase], Any]
        | Callable[[argparse.ArgumentParser], Any]
        | None
    ) = None,
    process_args: (
        Callable[[argparse.Namespace], argparse.Namespace | None] | None
    ) = None,
):
    def _cli1(function):
        module = inspect.getmodule(function)

        if inspect.iscoroutinefunction(function):

            @functools.wraps(function)
            async def _cli2(*args, **kwargs):
                with setup(function, add_arguments, process_args) as ba:
                    return await function(*ba.args, **ba.kwargs)

        else:

            @functools.wraps(function)
            def _cli2(*args, **kwargs):
                with setup(function, add_arguments, process_args) as ba:
                    return function(*ba.args, **ba.kwargs)

        _cli2.attributes = {
            "doc": function.__doc__ or module.__doc__ or "",
        }
        return _cli2

    return _cli1
